/********************************************

    cconfig -  config  library  for
     cserver
    David Marquez Hernandez  - 2016
    TFG UOC - Redes de Computadores
    Profesor  - Jesús Arribi Vilela

*********************************************/

#include "cconfig.h"
#include "../cerror/cerror.h"

/*
    NAME:   cconfig_init
    DESC:   Sets default configuration parameters
    INPUT:  Pointer to config struct, command line parameters.
    OUTPUT: 0 if success or error number if failure. Fills config struct with defaults.
    NOTES: CONFIG PRIORITY IS AS FOLLOWS:
        1) Command Line
        2) Default values
*/

int cconfig_init(cconfig *config,int argc, char **argv){
    int returnvalue=0;


    //Set default config values. Will be used if  command line is empty
    config->port=DEFAULTPORT;
    if (strcpy(config->www_root,DEFAULTWWWROOT)==NULL){
        cerror_show();
        return errno;
    };
    config->logging=DEFAULTLOGGING;
    logging=config->logging;
    config->max_concurrency=DEFAULTMAXCONCURRENCY;

    //Parse command line
    if ((returnvalue=cconfig_cmdline(config,argc,argv))!=0) {
        return returnvalue;
    }
    return 0;
}


/*
    NAME:   cconfig_cmdline
    DESC:   Reads command line parameters
    INPUT:  Pointer to config struct, command line parameters.
    OUTPUT: 0 if all ok, 1 if some parameter is not supported, 2 if help was requested by the user. If 0 fills the config struct.
    NOTES: Supported commands are:
        p port number
        r www root
        l logging (0/1/2)
        c max concurrency
*/
int cconfig_cmdline(cconfig *config, int argc, char **argv) {


    int index;
    int c;
    char errorstring[2048];
    int converted_integer;

    opterr = 0;
    while ((c = getopt (argc, argv, "p:r:l:c:h")) != -1)
    switch (c)
      {
      case 'p':
        converted_integer=atoi(optarg);
        if ((converted_integer>1) || (converted_integer<65535)){
            config->port=converted_integer;
        } else {
            sprintf(errorstring,"Port number must be between 1 and 65535: -%s",optarg);
            cerror_show_string(errorstring);
            cconfig_show_usage();
            return 1;
        }
        break;
      case 'r':
        if (strcpy(config->www_root,optarg)==NULL){
            cerror_show();
            return errno;
        };
        break;
      case 'l':
        converted_integer=atoi(optarg);
        if ((converted_integer==1) || (converted_integer==0) || (converted_integer==2)){
            config->logging = atoi(optarg);
            logging=config->logging;
        } else {
            sprintf(errorstring,"Logging must be 0,1 or 2: -%s",optarg);
            cerror_show_string(errorstring);
            cconfig_show_usage();
            return 1;
        }
        break;
      case 'c':
        converted_integer=atoi(optarg);
        if (converted_integer>0){
            config->max_concurrency= atoi(optarg);
        } else {
            sprintf(errorstring,"Max concurrency must be a positive number: -%s",optarg);
            cerror_show_string(errorstring);
            cconfig_show_usage();
            return 1;
        }
        break;
      case 'h':
        cconfig_show_usage();
        return 1;
        break;
      case '?':
        if ((optopt == 'p') || (optopt == 'r') || (optopt == 'l') || (optopt == 'c') || (optopt == 'f')) {
          sprintf(errorstring,"Missing argument for option: -%c",optopt);
        } else if (isprint (optopt)) {
          sprintf(errorstring,"Option not recognized:`-%c'.\n", optopt);
        } else{
          sprintf(errorstring,"Invalid character as option: `\\x%x'.\n", optopt);
        };
        cerror_show_string(errorstring);
        cconfig_show_usage();
        return 1;

      default:
        abort ();
    }

    for (index = optind; index < argc; index++)
    printf ("Arguments not valid: %s\n", argv[index]);

    return 0;
};



/*
    NAME:   cconfig_show_usage
    DESC:   Prints parameter help to screen
    INPUT:  Nothing.
    OUTPUT: Nothing.
*/


void cconfig_show_usage(void){

    printf ("CSERVER usage\n");
    printf("Configuration priority: 1) Command Line 2) Config File 3) Default parameters\n");
    printf("Parameters:\n");
    printf("-p port  : Port number that cserver will be listening on\n");
    printf("-r path  : Pathname to files that cserver will serve\n");
    printf("-l 0/1/2 : Logging level to syslog: 0 - Deactivated, 1 - Normal, 2 - Debugging\n");
    printf("-c num   : Sets the maximum concurrency limit\n");
    printf("-h       : This text\n");
    printf("Default parameters: Port: 1080 www_root: ./www/ max_concurrency: 1000 logging: 0\n");
}

/*
    NAME:   cconfig_show_config
    DESC:   Prints current config to screen.
    INPUT:  Config struct.
    OUTPUT: Nothing.
*/

void cconfig_show_config(cconfig *config){

    printf("Current Config is Port: %d, WWW-Root: %s, Logging: %d, Max Concurrency: %d\n",config->port,config->www_root,config->logging,config->max_concurrency);


}
