/********************************************

    cserver - concurrent web server
    David Marquez Hernandez  - 2016
    TFG UOC - Redes de Computadores
    Profesor  - Jesús Arribi Vilela

*********************************************/
#include <stdio.h>
#include <stdlib.h>
#include <arpa/inet.h>        //Internet addresses format conversion (inet_ntoa)
#include <netinet/ip.h>      //IP
#include <sys/socket.h>       // Sockets
#include <sys/types.h>        // pthreads, mutex, clock
#include <unistd.h>           // POSIX. getpid, fork, etc.
#include <sys/wait.h>         //waitpid for child processes
#include <pthread.h>          //thread management
#include <sys/resource.h>     //Stack size limit
#include <event.h>            //libevent

#include "../clogger/clogger.h"
#include "../cconfig/cconfig.h"
#include "../chttp/chttp.h"
#include "../cerror/cerror.h"

pthread_mutex_t count_mutex;


struct thread_arguments {
    cconfig config;
    int     *client_socket;
    char    *client_ip;
    int     *current_thread_number;
    struct event *event_socket_read;
};


/*
    NAME:   process_new_connection
    DESC:   Callback for socket read event. Reads request and sends the reply.
    INPUT:  server socket, event, pointer to arguments struct
    OUTPUT: nothing, per libevent convention must return void.
*/


void process_new_connection(int cserver_socket, short event, void *arguments){

    request_http_headers *request_headers;

    struct thread_arguments *args = arguments;

    if(logging==2)  printf("Read event for socket %d\n",*args->client_socket);


    request_headers=malloc(sizeof(request_http_headers));
    if (request_headers==NULL){
        cerror_show();
        close(*args->client_socket);
        event_del(args->event_socket_read);
        free(args->client_ip);
        free(args->client_socket);
        free(args->event_socket_read);
        free(args);

        return;
    }

    read_http_request(*args->client_socket,request_headers);
    reply_http_request(*args->client_socket, request_headers,&args->config,args->client_ip);

    close(*args->client_socket);

    event_del(args->event_socket_read);

    free(args->client_ip);
    free(args->client_socket);
    free(args->event_socket_read);
    free(args);
    free(request_headers);

}

/*
    NAME:   accept_callback
    DESC:   Callback for socket accept event.
    INPUT:  server socket, event, pointer to arguments struct
    OUTPUT: nothing, per libevent convention must return void.
*/


void accept_callback(int cserver_socket, short event, void *arguments){

    struct thread_arguments *args;
    struct sockaddr_in client_addr;
    socklen_t client_addr_size;
    int socket_flags;

    client_addr_size=sizeof(client_addr);

    cconfig *config=arguments;

    if (logging==2) printf("Connnection accepted callback running, config received port %d www root %s \n",config->port,config->www_root);


    args=malloc(sizeof(struct thread_arguments));
    if (args==NULL){
        cerror_show(); //Shows ERRNO error in both screen and logfile.
        clogger_stop();
        close(cserver_socket);
        return;
    }

    args->client_socket=malloc(sizeof(int));
    if (args->client_socket==NULL){
        cerror_show(); //Shows ERRNO error in both screen and logfile.
        clogger_stop();
        close(cserver_socket);
        free(args);
        return;
    }


    args->event_socket_read=calloc(1,sizeof(*args->event_socket_read));
    if (args->event_socket_read==NULL){
        cerror_show();
        clogger_stop();
        close(cserver_socket);
        return;
    }


    if ((*args->client_socket = accept(cserver_socket,(struct sockaddr *) &client_addr,&client_addr_size))<0) {
        cerror_show(); //Shows ERRNO error in both screen and logfile.
        clogger_stop();
        close(cserver_socket);
        free(args->client_socket);
        free(args);
        return;
    }


  //We now configure the socket as non-blocking, so the libevent main loop can iterate if there is now new connection
    socket_flags=fcntl(cserver_socket, F_GETFL); //Get current flags
    if (socket_flags<0) {
        cerror_show();
        clogger_stop();
        close(cserver_socket);
        return;
    };
    socket_flags |= O_NONBLOCK; //Add nonblock flag to whatever flags were set before.
    if (fcntl(cserver_socket,F_SETFL,socket_flags)<0) { //Set flags with nonblock
        cerror_show();
        clogger_stop();
        close(cserver_socket);
        return;
    }


    args->client_ip=malloc(INET_ADDRSTRLEN);
    if (args->client_ip==NULL){
        cerror_show(); //Shows ERRNO error in both screen and logfile.
        clogger_stop();
        close(cserver_socket);
        close(*args->client_socket);
        free(args->client_socket);
        free(args);
        return;
    }
    inet_ntop(AF_INET, &(client_addr.sin_addr), args->client_ip, INET_ADDRSTRLEN);

    if(logging==2) printf("Connected client IP is %s to client socket %d\n",args->client_ip,*args->client_socket);
    if(logging==2) printf("Setting read event...");

    args->config=*config;

    //We now configure an additional event for each read that needs to be done:

    //event_set(args->event_socket_read,*args->client_socket,EV_READ|EV_PERSIST,process_new_connection,args);
    event_set(args->event_socket_read,*args->client_socket,EV_READ,process_new_connection,args);
    if(logging==2) printf("STEP 1...");
    if (event_add(args->event_socket_read,NULL)<0) {
           cerror_show();
           clogger_stop();
           close(*args->client_socket);
           close(cserver_socket);
           free (args->client_socket);
           free (args->event_socket_read);
           free(args);
           return;
    }
    if(logging==2) printf("Done\n");

}


/*
    NAME:   main
    DESC:   Opens sockets, read config, controls concurrency and exectuion flow.
    INPUT:  Command line parameters.
    OUTPUT: 0 if successful, errno if failure.
*/


int main(int argc, char **argv) {

    cconfig config;
    int cserver_socket;
    int enable_reuse_addr,enable_reuse_port;
    struct sockaddr_in cserver_addr;
    int socket_flags;
    struct event event_accept; //event struct

    //Init logging:
        clogger_init();

    //Libevent init
        event_init();


    //Load config with this priority 1) CMD LINE 2) DEFAULT VALUES


        //Read config from command line and default values
        if (cconfig_init(&config, argc, argv)!=0) {
            clogger_stop();
            return errno;
        }
        //Print config to screen
        cconfig_show_config(&config);



    //Socket --> Listen
        //Create server socket
        if ((cserver_socket = socket (AF_INET,SOCK_STREAM,0))<0) { //Create a TCP (STREAM) socket.
            cerror_show(); //Shows ERRNO error in both screen and logfile.
            clogger_stop();
            return errno;
        }
        //Set reuse so if we restart the server very fast or the client dies it can bind again
        //Allows binding to a socket that is in TIME_WAIT state set by the kernel.
        enable_reuse_addr = 1;
        if (setsockopt(cserver_socket, SOL_SOCKET, SO_REUSEADDR, &enable_reuse_addr, sizeof(enable_reuse_addr)) < 0) {
            cerror_show(); //Shows ERRNO error in both screen and logfile.
            clogger_stop();
            close(cserver_socket);
            return errno;
        }

        #ifdef SO_REUSEPORT // Reuse socket for Linux 3.9+
            enable_reuse_port= 1;
            if (setsockopt(cserver_socket, SOL_SOCKET, SO_REUSEPORT, (const char*)&enable_reuse_port, sizeof(enable_reuse_port)) < 0){
                cerror_show(); //Shows ERRNO error in both screen and logfile.
                clogger_stop();
                close(cserver_socket);
                return errno;
            }
        #endif // SO_REUSEPORT
        //Fill server address and port
        cserver_addr.sin_family = AF_INET;             //IP protocol
        cserver_addr.sin_addr.s_addr = INADDR_ANY;     //Listen on all interfaces
        cserver_addr.sin_port = htons(config.port);   //Port from config
        //Assign address and port to the socket
        if (bind(cserver_socket,(struct sockaddr *) &cserver_addr,sizeof(cserver_addr))<0){
            cerror_show(); //Shows ERRNO error in both screen and logfile.
            clogger_stop();
            close(cserver_socket);
            return errno;
        }
        //Start listening with max_concurrency pending connections.
        //It is not ideal because we can have X connections active so we would have X+max_concurrency
        //But its a reasonable approach.
        if ((listen(cserver_socket,config.max_concurrency))<0) {
            cerror_show(); //Shows ERRNO error in both screen and logfile.
            clogger_stop();
            close(cserver_socket);
            return errno;
        }

        //We now configure the socket as non-blocking, so the libevent main loop can iterate if there is now new connection
        socket_flags=fcntl(cserver_socket, F_GETFL); //Get current flags
        if (socket_flags<0) {
            cerror_show();
            clogger_stop();
            close(cserver_socket);
            return errno;
        };
        socket_flags |= O_NONBLOCK; //Add nonblock flag to whatever flags were set before.
        if (fcntl(cserver_socket,F_SETFL,socket_flags)<0) { //Set flags with nonblock
            cerror_show();
            clogger_stop();
            close(cserver_socket);
            return errno;
        }

        if(logging==2) printf("CSERVER PID: %d PPID: %d  --> %s\n", getpid(),getppid(),"I'M PARENT. LISTEN SOCKET IS NOW OPEN");
        if(logging==2) printf("SOCKET %d set as non blocking, setting events...\n",cserver_socket);
        //Set event so callback function is called whenever an accept happens on the server socket.
        event_set(&event_accept,cserver_socket,EV_READ|EV_PERSIST,accept_callback,&config);

        if (event_add(&event_accept,NULL)<0) {
            cerror_show();
            clogger_stop();
            close(cserver_socket);
            return errno;
        }
        //Start event loop
        event_dispatch();

        //Clean exit:
        close(cserver_socket);
        clogger_stop();

        return 0;
}
