/********************************************

    cserver - concurrent web server
    David Marquez Hernandez  - 2016
    TFG UOC - Redes de Computadores
    Profesor  - Jesús Arribi Vilela

*********************************************/
#include <stdio.h>
#include <stdlib.h>
#include <arpa/inet.h>        //Internet addresses format conversion (inet_ntoa)
#include <netinet/ip.h>      //IP
#include <sys/socket.h>       // Sockets
#include <sys/types.h>        // pthreads, mutex, clock
#include <unistd.h>           // POSIX. getpid, fork, etc.
#include <sys/wait.h>         //waitpid for child processes

#include "../clogger/clogger.h"
#include "../cconfig/cconfig.h"
#include "../chttp/chttp.h"
#include "../cerror/cerror.h"


/*
    NAME:   process_new_connection
    DESC:   Main child process, reads request and sends the reply
    INPUT:  Config struct, server and client sockets, pointer to client ip string.
    OUTPUT: 0 if successful, errno if failure.
*/

int process_new_connection(cconfig config,int cserver_socket,int client_socket,char *client_ip) {
    request_http_headers request_headers;

    if (close(cserver_socket)<0) {
        cerror_show(); //Shows ERRNO error in both screen and logfile.
        clogger_stop();
        close(cserver_socket);
        close(client_socket);
        return errno;
    }
    if (logging==2) printf("CSERVER PID: %d PPID: %d-> %s\n", getpid(),getppid(),"I'M CHILD. NEW CLIENT CONNECTION RECEIVED");

    read_http_request(client_socket,&request_headers);
    reply_http_request(client_socket, &request_headers,&config,client_ip);


    close(client_socket);
    return 0;
}


/*
    NAME:   main
    DESC:   Opens sockets, read config, controls concurrency and exectuion flow.
    INPUT:  Command line parameters.
    OUTPUT: 0 if successful, errno if failure.
*/


int main(int argc, char **argv) {

    cconfig config;
    int cserver_socket, client_socket;
    int enable_reuse_addr,enable_reuse_port;
    struct sockaddr_in cserver_addr;
    struct sockaddr_in client_addr;
    char client_ip[INET_ADDRSTRLEN];
    socklen_t client_addr_size;
    int pid,pid_wait;
    int current_concurrency=0;

    //Init logging:
        clogger_init();


    //Load config with this priority 1) CMD LINE 2) DEFAULT VALUES


        //Read config from command line and default values
        if (cconfig_init(&config, argc, argv)!=0) {
            clogger_stop();
            return errno;
        }
        //Print config to screen
        cconfig_show_config(&config);


    //Socket --> Listen
        //Create server socket
        if ((cserver_socket = socket (AF_INET,SOCK_STREAM,0))<0) { //Create a TCP (STREAM) socket.
            cerror_show(); //Shows ERRNO error in both screen and logfile.
            clogger_stop();
            return errno;
        }
        //Set reuse so if we restart the server very fast or the client dies it can bind again
        //Allows binding to a socket that is in TIME_WAIT state set by the kernel.
        enable_reuse_addr = 1;
        if (setsockopt(cserver_socket, SOL_SOCKET, SO_REUSEADDR, &enable_reuse_addr, sizeof(enable_reuse_addr)) < 0) {
            cerror_show(); //Shows ERRNO error in both screen and logfile.
            clogger_stop();
            close(cserver_socket);
            return errno;
        }

        #ifdef SO_REUSEPORT // Reuse socket for Linux 3.9+
            enable_reuse_port= 1;
            if (setsockopt(cserver_socket, SOL_SOCKET, SO_REUSEPORT, (const char*)&enable_reuse_port, sizeof(enable_reuse_port)) < 0){
                cerror_show(); //Shows ERRNO error in both screen and logfile.
                clogger_stop();
                close(cserver_socket);
                return errno;
            }
        #endif // SO_REUSEPORT
        //Fill server address and port
        cserver_addr.sin_family = AF_INET;             //IP protocol
        cserver_addr.sin_addr.s_addr = INADDR_ANY;     //Listen on all interfaces
        cserver_addr.sin_port = htons(config.port);   //Port from config
        //Assign address and port to the socket
        if (bind(cserver_socket,(struct sockaddr *) &cserver_addr,sizeof(cserver_addr))<0){
            cerror_show(); //Shows ERRNO error in both screen and logfile.
            clogger_stop();
            close(cserver_socket);
            return errno;
        }
        //Start listening with max_concurrency pending connections.
        //It is not ideal because we can have X connections active so we would have X+max_concurrency
        //But its a reasonable approach.
        if ((listen(cserver_socket,config.max_concurrency))<0) {
            cerror_show(); //Shows ERRNO error in both screen and logfile.
            clogger_stop();
            close(cserver_socket);
            return errno;
        }
        if (logging==2) printf("CSERVER PID: %d PPID: %d  --> %s\n", getpid(),getppid(),"I'M PARENT. LISTEN SOCKET IS NOW OPEN");


        while (1) { //Loop for connections

            //New connection arrives
            client_addr_size=sizeof(client_addr);
            if ((client_socket = accept(cserver_socket,(struct sockaddr *) &client_addr,&client_addr_size))<0) {
                cerror_show(); //Shows ERRNO error in both screen and logfile.
                clogger_stop();
                close(cserver_socket);
                return errno;
            }
            inet_ntop(AF_INET, &(client_addr.sin_addr), client_ip, INET_ADDRSTRLEN);

            if (logging==2) printf("Connected client IP is %s\n",client_ip);

            //We have a new connection, spawn a new child process:
            //IF and only if the current process count is < max_concurrency, if not, wait for a child to finish.

            if (current_concurrency<config.max_concurrency){
                current_concurrency++;
                if (logging==2) printf("CSERVER current con %d max con %d\n",current_concurrency,config.max_concurrency);
                pid=fork();
                if (pid==0) { //fork returns 0 to child, pid to parent.
                    //CHILD
                    process_new_connection(config,cserver_socket,client_socket,(char*)&client_ip);
                    clogger_stop();
                    exit(EXIT_SUCCESS); //CHILD PROCESS TERMINATES HERE.

                } else if (pid==-1){ //PARENT ERROR FORKING
                       cerror_show(); //Shows ERRNO error in both screen and logfile.
                       clogger_stop();
                       close(cserver_socket);
                       close(client_socket);
                       return errno;
                }
            }

            if (logging==2) printf("CSERVER PID: %d PPID: %d I'M PARENT WAITING FOR A CHILD TO FINISH... %s\n", getpid(),getppid(),".");

            close(client_socket);
            if ((pid_wait=waitpid(-1, NULL, WNOHANG))<0){ //Returns -1 if error, 0 if no child has changed state and PID number if child PID has finished.
                cerror_show(); //Shows ERRNO error in both screen and logfile.
                clogger_stop();
                close(cserver_socket);
                close(client_socket);
                return errno;
            };
            if (pid_wait>1) {  //A child has finished, we can spawn a new one
                if (logging==2) printf("CSERVER PID: %d PPID: %d I'M PARENT. CHILD WITH PID: %d HAS FINISHED!%s\n", getpid(),getppid(),pid_wait,".");
                current_concurrency--; //we can spawn a new children
            }


        }

        //Clean exit:
        close(cserver_socket);
        close(client_socket);
        clogger_stop();

        return 0;
}
