/********************************************

    cserver - concurrent web server
    David Marquez Hernandez  - 2016
    TFG UOC - Redes de Computadores
    Profesor  - Jesús Arribi Vilela

*********************************************/
#include <stdio.h>
#include <stdlib.h>
#include <arpa/inet.h>        //Internet addresses format conversion (inet_ntoa)
#include <netinet/ip.h>      //IP
#include <sys/socket.h>       // Sockets
#include <sys/types.h>        // pthreads, mutex, clock
#include <unistd.h>           // POSIX. getpid, fork, etc.
#include <sys/wait.h>         //waitpid for child processes
#include <pthread.h>          //thread management
#include <sys/resource.h>     //Stack size limit

#include "../clogger/clogger.h"
#include "../cconfig/cconfig.h"
#include "../chttp/chttp.h"
#include "../cerror/cerror.h"

pthread_mutex_t count_mutex;


struct thread_arguments {
    cconfig config;
    int     *client_socket;
    char    *client_ip;
    int     *current_thread_number;
};

/*
    NAME:   process_new_connection
    DESC:   Main child process, reads request and sends the reply
    INPUT:  Config struct, server and client sockets, pointer to client ip string.
    OUTPUT: 0 if successful, errno if failure.
*/

//int process_new_connection(cconfig config,int cserver_socket,int client_socket,char *client_ip,int *current_thread_number) {

void* process_new_connection(void *arguments) {
    request_http_headers *request_headers;

    struct thread_arguments *args = arguments;


    if(logging==2) printf("This is thread ID %ld connected to IP %s to client socket %d \n",(long int)pthread_self(),args->client_ip,*args->client_socket);

    request_headers=malloc(sizeof(request_http_headers));
    if (request_headers==NULL){
        cerror_show();
        close(*args->client_socket);
        free(args->client_ip);
        free(args->client_socket);
        free(args);
        return 0;
    }

    read_http_request(*args->client_socket,request_headers);
    reply_http_request(*args->client_socket, request_headers,&args->config,args->client_ip);


    close(*args->client_socket);

    if(logging==2)  printf("Current thread number is %ld ",(long int)*args->current_thread_number);

    pthread_mutex_lock(&count_mutex);
        *args->current_thread_number=*args->current_thread_number-1; //From the thread, just before finishing, reduce the current  number of threads.
    pthread_mutex_unlock(&count_mutex);

    if(logging==2)  printf("Decreasing  to %ld\n",(long int)*args->current_thread_number);

    //Free client ip, whose memory was reserved in main function
    free(args->client_ip);
    free(args->client_socket);
    free(args);
    free(request_headers);
    return 0;
}

/*   NAME:   main
    DESC:   Opens sockets, read config, controls concurrency and exectuion flow.
    INPUT:  Command line parameters.
    OUTPUT: 0 if successful, errno if failure.
*/


int main(int argc, char **argv) {

    cconfig config;
    int cserver_socket;
    int enable_reuse_addr,enable_reuse_port;
    struct sockaddr_in cserver_addr;
    struct sockaddr_in client_addr;
    socklen_t client_addr_size;
    int current_concurrency=0;
    pthread_t thread;
    pthread_attr_t thread_attributes;
    struct thread_arguments *args;

    //Init logging:
        clogger_init();


    //Load config with this priority 1) CMD LINE 2) DEFAULT VALUES


        //Read config from command line and default values
        if (cconfig_init(&config, argc, argv)!=0) {
            clogger_stop();
            return errno;
        }
        //Print config to screen
        cconfig_show_config(&config);



    //Socket --> Listen
        //Create server socket
        if ((cserver_socket = socket (AF_INET,SOCK_STREAM,0))<0) { //Create a TCP (STREAM) socket.
            cerror_show(); //Shows ERRNO error in both screen and logfile.
            clogger_stop();
            return errno;
        }
        //Set reuse so if we restart the server very fast or the client dies it can bind again
        //Allows binding to a socket that is in TIME_WAIT state set by the kernel.
        enable_reuse_addr = 1;
        if (setsockopt(cserver_socket, SOL_SOCKET, SO_REUSEADDR, &enable_reuse_addr, sizeof(enable_reuse_addr)) < 0) {
            cerror_show(); //Shows ERRNO error in both screen and logfile.
            clogger_stop();
            close(cserver_socket);
            return errno;
        }

        #ifdef SO_REUSEPORT // Reuse socket for Linux 3.9+
            enable_reuse_port= 1;
            if (setsockopt(cserver_socket, SOL_SOCKET, SO_REUSEPORT, (const char*)&enable_reuse_port, sizeof(enable_reuse_port)) < 0){
                cerror_show(); //Shows ERRNO error in both screen and logfile.
                clogger_stop();
                close(cserver_socket);
                return errno;
            }
        #endif // SO_REUSEPORT
        //Fill server address and port
        cserver_addr.sin_family = AF_INET;             //IP protocol
        cserver_addr.sin_addr.s_addr = INADDR_ANY;     //Listen on all interfaces
        cserver_addr.sin_port = htons(config.port);   //Port from config
        //Assign address and port to the socket
        if (bind(cserver_socket,(struct sockaddr *) &cserver_addr,sizeof(cserver_addr))<0){
            cerror_show(); //Shows ERRNO error in both screen and logfile.
            clogger_stop();
            close(cserver_socket);
            return errno;
        }
        //Start listening with max_concurrency pending connections.
        //It is not ideal because we can have X connections active so we would have X+max_concurrency
        //But its a reasonable approach.
        if ((listen(cserver_socket,config.max_concurrency))<0) {
            cerror_show(); //Shows ERRNO error in both screen and logfile.
            clogger_stop();
            close(cserver_socket);
            return errno;
        }
        if(logging==2) printf("CSERVER PID: %d PPID: %d  --> %s\n", getpid(),getppid(),"I'M PARENT. LISTEN SOCKET IS NOW OPEN");
        if(logging==2) printf("CSERVER parent thread ID is %ld\n",(long int)pthread_self());

        client_addr_size=sizeof(client_addr);

        //Set thread attributes: detached and increased stack size.
        pthread_attr_init(&thread_attributes);
        pthread_attr_setdetachstate(&thread_attributes, PTHREAD_CREATE_DETACHED);
        pthread_attr_setstacksize(&thread_attributes,131072);
        while (1) { //Loop for connections

            args=malloc(sizeof(struct thread_arguments));
            if (args==NULL){
                cerror_show(); //Shows ERRNO error in both screen and logfile.
                clogger_stop();
                close(cserver_socket);
                return errno;
            }

            args->client_socket=malloc(sizeof(int));
            if (args->client_socket==NULL){
                cerror_show(); //Shows ERRNO error in both screen and logfile.
                clogger_stop();
                close(cserver_socket);
                free(args);
                return errno;
            }
            if ((*args->client_socket = accept(cserver_socket,(struct sockaddr *) &client_addr,&client_addr_size))<0) {
                cerror_show(); //Shows ERRNO error in both screen and logfile.
                clogger_stop();
                close(cserver_socket);
                free(args->client_socket);
                free(args);
                return errno;
            }
            args->client_ip=malloc(INET_ADDRSTRLEN);
            if (args->client_ip==NULL){
                cerror_show(); //Shows ERRNO error in both screen and logfile.
                clogger_stop();
                close(cserver_socket);
                close(*args->client_socket);
                free(args->client_socket);
                free(args);
                return errno;
            }
            inet_ntop(AF_INET, &(client_addr.sin_addr), args->client_ip, INET_ADDRSTRLEN);

            if(logging==2) printf("Connected client IP is %s to client socket %d\n",args->client_ip,*args->client_socket);

            //We have a new connection, spawn a new thread
            //IF and only if the current threadcount is < max_concurrency, if not, keep waiting

            if (current_concurrency<config.max_concurrency){
                pthread_mutex_lock(&count_mutex);
                    current_concurrency++;
                pthread_mutex_unlock(&count_mutex);
                if(logging==2)  printf("CSERVER current con %d max con %d\n",current_concurrency,config.max_concurrency);

                //Fill thread arguments struct
                args->config=config;
                args->current_thread_number=&current_concurrency;

                if ((thread=pthread_create(&thread,&thread_attributes,process_new_connection,(void *)args))<0) {
                    cerror_show(); //Shows ERRNO error in both screen and logfile.
                    clogger_stop();
                    close(cserver_socket);
                    close(*args->client_socket);
                    free(args->client_socket);
                    free(args->client_ip);
                    free(args);
                    return errno;
                }

            }

        }

        //Clean exit:
        close(cserver_socket);
        clogger_stop();
        free(args->client_socket);
        free(args->client_ip);
        free(args);
        pthread_exit(NULL);
        return 0;
}
